do	-- Globals

	EXPIRY = os.time {year=2037, month=1, day=1, hour=0}

	EX_CMD = {}
	PRX_CMD = {}
	NOTIFY = {}
	DEV_MSG = {}
	LUA_ACTION = {}
	Helper = {}
	RESPONSES = {}
	Camera = {}
	URL = {}

	--[[
	Tables of functions
	The following tables are function containers that are called within the following functions:

		OnDriverInit()
			- first calls all functions contained within ON_DRIVER_EARLY_INIT table
			- then calls all functions contained within ON_DRIVER_INIT table
		OnDriverLateInit()
			- calls all functions contained within ON_DRIVER_LATEINIT table
		OnDriverUpdate()
			- calls all functions contained within ON_DRIVER_UPDATE table
		OnDriverDestroyed()
			- calls all functions contained within ON_DRIVER_DESTROYED table
		OnPropertyChanged()
			- calls all functions contained within ON_PROPERTY_CHANGED table
	--]]
	ON_DRIVER_INIT = {}
	ON_DRIVER_EARLY_INIT = {}
	ON_DRIVER_LATEINIT = {}
	ON_DRIVER_UPDATE = {}
	ON_DRIVER_DESTROYED = {}
	ON_PROPERTY_CHANGED = {}

	-- Constants
	DEFAULT_PROXY_BINDINGID = 5001

	--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
	-- Camera Proxy Code - This section contains the methods that must be
	--   modified/implemented to create a driver for a specific camera.
	--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

	-- Global Variables
	gVersion = "1.0.1" -- variable to track major, minor, revision.

	-- Global Variables - Initialized, updated, and persisted by proxy driver
	gAddress = "localhost"
	gHttpPort = 80
	gRtspPort = 80
	gAuthRequired = true
	gAauthType = "BASIC" -- "BASIC" or "DIGEST"
	gUsername = "admin"
	gPassword = "admin"
	gPubliclyAccessible = false

	-- Global Variables - Initialized, updated, and persisted by lua driver
	gCameraID = 0
	gPtzIncrement = 7
	gPTZTime = 250
	gMotionPoll = 10
	gStatusPrint = false

	gImageValues = {} -- brightness etc
	gImageValuesMin = 0
	gImageValuesMax = 32

	gDetectionState = {}
	gDetectionBinding = {}
	gDwell = {}

	gContactBound = {}
	gCanPoll = false
	gServerPort = nil

	--Global Tables
	UI_REQ = {}
	gTickets = {}
end

--[[
	This will be called by the base template code from its OnDriverInit function.
]]--
function ON_DRIVER_EARLY_INIT.Camera()
	-- Recreate bindings
	if (PersistData == nil) then
		PersistData = {}
	end

	if(PersistData["detection"] == nil) then
		PersistData["detection"] = {}
		SETTINGUP = true -- if no persistent data this will cause contacts to be created on discovery.
	end

	if(PersistData ~= nil) then
		for k,v in pairs(PersistData["detection"]) do
			C4:AddDynamicBinding(k, "CONTROL", true, v, "CONTACT_SENSOR", false, false)
			print("Added Detection: " .. k .. " with value: " .. v)
		end
	end
end

function ON_DRIVER_INIT.Camera()
	C4:UpdateProperty("Version", gVersion)
	gInitTimer = Timer:Create("Init", 5, "SECONDS", OnInitTimerExpired)
	gPTZTimer = Timer:Create("PTZ", gPTZTime, "MILLISECONDS", OnPTZTimerExpired)
	gSettingsTimer = Timer:Create("Settings", 5, "SECONDS", OnSettingsTimerExpired)
	gResetInfoTimer = Timer:Create("Settings", 30, "SECONDS", OnResetInfoTimerExpired)
	gSyncTimer = Timer:Create("Time Sync", 24, "HOURS", OnSyncTimerExpired, true)
	gMotionTimer = Timer:Create("Motion", gMotionPoll, "SECONDS", OnMotionTimerExpired, true)
	gRebootTimer = Timer:Create("Motion", 60, "SECONDS", OnRebootTimerExpired)
	gStatusTimer = Timer:Create("Motion", 180, "SECONDS", OnStatusTimerExpired) -- checks camera every 180 seconds for settings changes
	gStatusPrint = true

	Dbg:Trace("ON_DRIVER_INIT.Camera")

end


function ON_DRIVER_LATEINIT.Camera()
	gServerPort = math.random (49152, 65535)
	C4:CreateServer (gServerPort)

	local property_defaults = {}
	property_defaults.HTTP_PORT = C4:GetCapability("default_http_port") or 80
	property_defaults.RTSP_PORT = C4:GetCapability("default_rtsp_port") or 554
	property_defaults.AUTHENTICATION_REQUIRED = C4:GetCapability("default_authentication_required") or false
	property_defaults.AUTHENTICATION_TYPE = C4:GetCapability("default_authentication_type") or "BASIC"
	property_defaults.USERNAME = C4:GetCapability("default_username") or "admin"
	property_defaults.PASSWORD = C4:GetCapability("default_password") or "admin"

	C4:SendToProxy(DEFAULT_PROXY_BINDINGID, "PROPERTY_DEFAULTS", property_defaults, "NOTIFY")

	Camera.GetSettings()

end
--[[
	This will be called by the base template code from its OnDriverUpdate function.
]]--
function ON_DRIVER_UPDATE.Camera()
	OnDriverInit()
end

----------------------------------------------------------------------------
-- Define property handlers for properties added by this protocol driver
-- Camera ID -> CameraID
-- PTZ Increment -> PTZIncrement
----------------------------------------------------------------------------

function ON_PROPERTY_CHANGED.CameraID(propertyValue)
	Dbg:Trace("ON_PROPERTY.CameraID: " .. propertyValue)
	gCameraID = tonumber(propertyValue)
	gSettingsTimer:StartTimer()
end

function ON_PROPERTY_CHANGED.PanTiltZoomSpeed(propertyValue)
	Dbg:Trace("ON_PROPERTY.PanTiltZoomSpeed: " .. propertyValue)

	gPtzIncrement = tonumber(propertyValue)
end

function ON_PROPERTY_CHANGED.PanTiltZoomTime(propertyValue)
	Dbg:Trace("ON_PROPERTY.PanTiltZoomTime: " .. propertyValue)
	gPTZTime = tonumber(propertyValue)
end

function ON_PROPERTY_CHANGED.Pollformotioninseconds(propertyValue)
	if (propertyValue == 'Off') then
		gMotionTimer:KillTimer()
		gMotionPoll = 0
	else
		gMotionPoll = tonumber(propertyValue)
		local url = URL.BuildHTTP('SetMotionDetect.cgi?CH=' .. gCameraID .. '&Mot_Interval=' .. gMotionPoll) -- write as Mot_Interval read back as Interval
		print (url)
		local ticket = URL.GetRequest(url)
		gTickets[ticket] = 'SetAlarmDuration'
	end
end

function ON_PROPERTY_CHANGED.SyncwithC4Time(propertyValue)
	Dbg:Trace("ON_PROPERTY.SyncTime: " .. propertyValue)
	if (propertyValue == "Yes") then
		gSyncTimer:StartTimer()
		LUA_ACTION.Time ()
	else
		gSyncTimer:KillTimer()
	end
end

--[[
	This function is called when a UI (Navigator) requests data.
--]]
function UIRequest(sRequest, tParams)
	if (Helper.Unlicenced ()) then return end
	local ret = ""

	if (sRequest ~= nil) then
		tParams = tParams or {}   -- initial table variable if nil
		Dbg:Trace("UIRequest(): " .. sRequest .. "; Call Function UI_REQ." .. sRequest .. "()")
		Dbg:Info(tParams)

		if (UI_REQ[sRequest]) ~= nil then
			ret = UI_REQ[sRequest](tParams)
		else
			Dbg:Alert("UIRequest: Unhandled request = " .. sRequest)
		end
	end

	return ret
end

--[[
	Return the query string required for an HTTP snapshot URL request.
--]]
function UI_REQ.GET_SNAPSHOT_QUERY_STRING(tParams)
	tParams = tParams or {}
    local size_x = tonumber(tParams["SIZE_X"]) or 640
    local size_y = tonumber(tParams["SIZE_Y"]) or 480
    Dbg:Trace("UI_REQ.GET_SNAPSHOT_QUERY_STRING (SIZE_X: " .. size_x .. " SIZE_Y: " .. size_y .. ")")
	
	snapshot_query_string = "API/OnvifSnapshot/Get"

	return "<snapshot_query_string>" .. C4:XmlEscapeString(snapshot_query_string) .. "</snapshot_query_string>"
end

--[[
	Return the query string required for an HTTP image push URL request.
--]]
function UI_REQ.GET_MJPEG_QUERY_STRING(tParams)
	tParams = tParams or {}
    local size_x = tonumber(tParams["SIZE_X"]) or 640
    local size_y = tonumber(tParams["SIZE_Y"]) or 480
	local delay = tonumber(tParams["DELAY"]) or 200
    Dbg:Trace("UI_REQ.GET_MJPEG_QUERY_STRING (SIZE_X: " .. size_x .. " SIZE_Y: " .. size_y .. " DELAY: " .. delay .. ")")
	local chmun = Properties['Channel'] - 1
	mjpeg_query_string = "mjpeg?channel=" .. chmun

	return "<mjpeg_query_string>" .. C4:XmlEscapeString(mjpeg_query_string) .. "</mjpeg_query_string>"
end

--[[
	Return the query string required to establish Rtsp connection. May be empty string.
--]]
function UI_REQ.GET_RTSP_H264_QUERY_STRING(tParams)
	tParams = tParams or {}
    local size_x = tonumber(tParams["SIZE_X"]) or 640
    local size_y = tonumber(tParams["SIZE_Y"]) or 480
	local delay = tonumber(tParams["DELAY"]) or 50
    Dbg:Trace("UI_REQ.GET_RTSP_H264_QUERY_STRING (SIZE_X: " .. size_x .. " SIZE_Y: " .. size_y .. " DELAY: " .. delay .. ")")

	--rtsp_query_string = gDiscoveredPath1
	local rtsp_query_string = 'rtsp/streaming?channel=' .. Properties['Channel'] .. '&subtype=' .. Properties['Stream Type']

	return "<rtsp_h264_query_string>" .. C4:XmlEscapeString(rtsp_query_string) .. "</rtsp_h264_query_string>"
end


--[[
	Implement the following commands as required by capabilities.
	The function 'URL.GetRequest(url)' can be used to perform HTTP Get.
	example:
		local url = URL.BuildHTTP("sample/queryString?parameter=value&parameter2=value2")
		local ticket = URL.GetRequest(url)
--]]

function PRX_CMD.PAN_LEFT(idBinding, tParams)
	Dbg:Trace("PRX_CMD.PAN_LEFT")
	local url = URL.BuildHTTP(Camera.PTZControl("Left"))
	local ticket = URL.GetRequest(url)
	Dbg:Trace("PRX_CMD.PAN_LEFT: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.PAN_RIGHT(idBinding, tParams)
	Dbg:Trace("PRX_CMD.PAN_RIGHT")

	local url = URL.BuildHTTP(Camera.PTZControl("Right"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.PAN_RIGHT: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.PAN_SCAN(idBinding, tParams) -- Not supported
	Dbg:Trace("PRX_CMD.PAN_SCAN")

	-- Not supported
end

function PRX_CMD.TILT_UP(idBinding, tParams)
	Dbg:Trace("PRX_CMD.TILT_UP")

	local url = URL.BuildHTTP(Camera.PTZControl("Up"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.TILT_UP: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.TILT_DOWN(idBinding, tParams)
	Dbg:Trace("PRX_CMD.TILT_DOWN")

	local url = URL.BuildHTTP(Camera.PTZControl("Down"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.TILT_DOWN: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.TILT_SCAN(idBinding, tParams) -- Not supported
	Dbg:Trace("PRX_CMD.TILT_SCAN")

	-- Not supported
end

function PRX_CMD.ZOOM_IN(idBinding, tParams)
	Dbg:Trace("PRX_CMD.ZOOM_IN")

	local url = URL.BuildHTTP(Camera.PTZControl("ZoomIn"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.ZOOM_IN: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.ZOOM_OUT(idBinding, tParams)
	Dbg:Trace("PRX_CMD.ZOOM_OUT")

	local url = URL.BuildHTTP(Camera.PTZControl("ZoomOut"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.ZOOM_OUT: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.FOCUS_NEAR(idBinding, tParams)
	Dbg:Trace("PRX_CMD.FOCUS_NEAR")

	local url = URL.BuildHTTP(Camera.PTZControl("FocusNear"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.FOCUS_NEAR: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.FOCUS_FAR(idBinding, tParams)
	Dbg:Trace("PRX_CMD.FOCUS_FAR")

	local url = URL.BuildHTTP(Camera.PTZControl("FocusFar"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.FOCUS_FAR: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.IRIS_OPEN(idBinding, tParams)
	Dbg:Trace("PRX_CMD.IRIS_OPEN")

	local url = URL.BuildHTTP(Camera.PTZControl("IRISLarge"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.IRIS_OPEN: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.IRIS_CLOSE(idBinding, tParams)
	Dbg:Trace("PRX_CMD.IRIS_CLOSE")

	local url = URL.BuildHTTP(Camera.PTZControl("IRISSmall"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.IRIS_CLOSE: Ticket = " .. ticket .. " URL = " .. url)
end
function PRX_CMD.HOME(idBinding, tParams)
	Dbg:Trace("PRX_CMD.HOME")

	local url = URL.BuildHTTP(Camera.PTZControl("Home"))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.HOME: Ticket = " .. ticket .. " URL = " .. url)
end

function PRX_CMD.MOVE_TO(idBinding, tParams)
	tParams = tParams or {}
    local width = tonumber(tParams["WIDTH"]) or 0
    local height = tonumber(tParams["HEIGHT"]) or 0
	local x_index = tonumber(tParams["X_INDEX"]) or 0
	local y_index = tonumber(tParams["Y_INDEX"]) or 0

    Dbg:Trace("PRX_CMD.MOVE_TO (WIDTH: " .. width .. " HEIGHT: " .. height .. " X_INDEX: " .. x_index .." Y_INDEX: " .. y_index .. ")")

	x_index = (x_index - (width / 2)) / (width / 2)
	y_index = ((height / 2) - y_index) / (height/ 2)

	local cmd = nil
	if ((x_index < 0) and (y_index < 0)) then
		cmd = 'DownLeft'
	elseif ((x_index > 0) and (y_index < 0)) then
		cmd = 'DownRight'
	elseif ((x_index > 0) and (y_index > 0)) then
		cmd = 'UpRight'
	elseif ((x_index < 0) and (y_index > 0)) then
		cmd = 'UpLeft'
	end

	if (cmd) then
		local url = URL.BuildHTTP(Camera.PTZControl(cmd))
		local ticket = URL.GetRequest(url)
	end
end

function PRX_CMD.NIGHT_MODE(idBinding, tParams) -- To do
	tParams = tParams or {}
    local enable = Helper.ToBoolean(tParams["ENABLE"])
    Dbg:Trace("PRX_CMD.NIGHT_MODE (ENABLE: " .. enable .. ")")

	-- TODO: implement
end

function PRX_CMD.PRESET(idBinding, tParams)
	tParams = tParams or {}
	local index = tonumber(tParams["INDEX"]) or 1
    Dbg:Trace("PRX_CMD.PRESET (INDEX: " .. index .. ")")

	local url = URL.BuildHTTP(Camera.GotoPreset(index))
	local ticket = URL.GetRequest(url)

	Dbg:Trace("PRX_CMD.PRESET: Ticket = " .. ticket .. " URL = " .. url)
end

--[[
	The "SET" commands are called by proxy to initialize and update the
	global variables used by the protocol driver. No code changes to the
	"SET" commands is required.
--]]

function PRX_CMD.SET_ADDRESS(idBinding, tParams)
	tParams = tParams or {}
	gAddress = tParams["ADDRESS"] or gAddress
	gSettingsTimer:StartTimer()
    Dbg:Trace("PRX_CMD.SET_ADDRESS (ADDRESS: " .. gAddress .. ")")
end

function PRX_CMD.SET_HTTP_PORT(idBinding, tParams)
	tParams = tParams or {}
	gHttpPort = tonumber(tParams["PORT"]) or gHttpPort
	gSettingsTimer:StartTimer()
    Dbg:Trace("PRX_CMD.SET_HTTP_PORT (PORT: " .. gHttpPort .. ")")
end

function PRX_CMD.SET_RTSP_PORT(idBinding, tParams)
	tParams = tParams or {}
	gRtspPort = tonumber(tParams["PORT"]) or gRtspPort
	gSettingsTimer:StartTimer()
    Dbg:Trace("PRX_CMD.SET_RTSP_PORT (PORT: " .. gRtspPort .. ")")
end

function PRX_CMD.SET_AUTHENTICATION_REQUIRED(idBinding, tParams)
	tParams = tParams or {}
	gAuthRequired = Helper.ToBoolean(tParams["REQUIRED"])
	gSettingsTimer:StartTimer()
	Dbg:Trace("PRX_CMD.SET_AUTHENTICATION_REQUIRED (REQUIRED: " .. tostring(gAuthRequired) .. ")")
end

function PRX_CMD.SET_AUTHENTICATION_TYPE(idBinding, tParams)
	tParams = tParams or {}
	gAauthType = tParams["TYPE"] or gAauthType
	gSettingsTimer:StartTimer()
    Dbg:Trace("PRX_CMD.SET_AUTHENTICATION_TYPE (TYPE: " .. gAauthType .. ")")
end

function PRX_CMD.SET_USERNAME(idBinding, tParams)
	tParams = tParams or {}
	gUsername = tParams["USERNAME"] or gUsername
	gSettingsTimer:StartTimer()
    Dbg:Trace("PRX_CMD.SET_USERNAME (USERNAME: " .. gUsername .. ")")
end

function PRX_CMD.SET_PASSWORD(idBinding, tParams)
	tParams = tParams or {}
	gPassword = tParams["PASSWORD"] or gPassword
	gSettingsTimer:StartTimer()
    Dbg:Trace("PRX_CMD.SET_PASSWORD (PASSWORD: " .. gPassword .. ")")
end

function PRX_CMD.SET_PUBLICLY_ACCESSIBLE(idBinding, tParams)
	tParams = tParams or {}
	gPubliclyAccessible = Helper.ToBoolean(tParams["PUBLICLY_ACCESSIBLE"])
    Dbg:Trace("PRX_CMD.SET_PUBLICLY_ACCESSIBLE (PUBLICLY_ACCESSIBLE: " .. gPubliclyAccessible .. ")")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Callback functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

--[[
	The ReceivedAsync function is called in response to 'URL.GetRequest'.
	The ticketId is the number returned from the request.
--]]
function ReceivedAsync(ticketId, strData, responseCode, tHeaders, strError)
	strData = strData or ""
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}

	--Dbg:Trace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(strData))

	local data_length = string.len(strData)
	if (data_length > 128) then data_length = 128 end
	Dbg:Trace("ReceivedAsync[" .. ticketId .. "]: Data: " .. string.sub(strData, 1, data_length))

	print("respcode:" .. responseCode .. "ticketId:" .. ticketId)

	print("respdata:" .. strData .. "strerr:" .. strError)
	if (gTickets[ticketId] ~= nil and type(RESPONSES[gTickets[ticketId]]) == "function") then -- Need to parse response
		RESPONSES[gTickets[ticketId]](strData, responseCode, tHeaders)
		gTickets[ticketId] = nil -- release space ??
	end

	if (responseCode == 0) then
		if (gTickets[ticketId] ~= nil) then
			if (string.find(gTickets[ticketId], 'ServerSettings')) then
				RESPONSES[gTickets[ticketId]]('Not Supported', responseCode, tHeaders)
				return
			end
		end
		--PrintTable(tHeaders, ">    ")
		--print (strData)
	end


end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Response functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

function RESPONSES.Login_first(data,responseCode, tHeaders)
	PrintTable(tHeaders, ">    ")
	print (data)
	print(responseCode)
end

function RESPONSES.GetCameraModel(data,responseCode, tHeaders)
	gModel = string.match(data,'CameraMode=(.-)<br>')
	if (gModel == nil) then
		Helper.StatusPrint ('No response to camera model')
		gError = true
		return
	end
	Helper.StatusPrint ('Camera Model ' .. gModel)
	C4:UpdateProperty ('Camera Model', gModel)
end

function RESPONSES.GetCameraName(data,responseCode, tHeaders)
	local name = string.match(data,'Name=(.-)<br>')
	if (name == nil) then
		Helper.StatusPrint ('No response to camera name, check network and port settings')
		gError = true
		return
	end
	Helper.StatusPrint ('Camera Name ' .. name)
	C4:UpdateProperty ('Camera Name', name)
end

function RESPONSES.GetCameraType(data,responseCode, tHeaders)
	local push = string.match(data,'Server_push_over_http=(%d)<br>')
	if (not push) then
		Helper.StatusPrint('No response to camera type query')
		gError = true
		return
	end
	if (push ~= '1') then -- does not support MJPEG push
		Helper.StatusPrint ('Server push is not supported - this camera is not Control4 compatible')
		gError = true
	end
	if (string.match(data,'Ptz=(%d)<br>') == '1') then -- does not support PTZ
		Helper.StatusPrint ('WRONG DRIVER - The camera has PTZ')
		gError = true
	end
	local codecs = string.lower(string.match(data,'CodecType1=(.-)<br>'))
	codecs = codecs .. string.lower(string.match(data,'CodecType2=(.-)<br>'))
	if (not string.find (codecs, 'mjpeg')) then -- does not support MJPEG
		Helper.StatusPrint ('MJPEG is not supported - this camera is not Control4 compatible')
		gError = true
	end
end

function RESPONSES.GetEventSettings (data,responseCode, tHeaders)
	local tamper = string.match(data,'Tampering_En=(%d)<br>')
	if (not tamper) then -- does not support tamper
		return
	end
	if (tamper == '1') then -- tamper enabled
		-- TO DO -- Set Dynamic binding and persist
	else
		Helper.StatusPrint ('Camera supports tamper detect but this is not enabled, virtual contact not created')
		gError = true
	end
end

function RESPONSES.GetImageSettings (data,responseCode, tHeaders)
	if (string.find (data, "Failed to get")) then
		Helper.StatusPrint ("ERROR: No repsonse to Image settings query")
		return
	end
	for name, value in string.gmatch (data, "(%w+)=([+-]?%d+)<br>") do
		gImageValues[name] = tonumber(value)
		if (value == '-1') then
			Helper.StatusPrint ("Adjustment of " .. name .. " is not supported.")
		else
			Helper.StatusPrint ("Adjustment of " .. name .. " is supported.")
		end
	end

	if gImageValues['Orientation'] then
		if (gImageValues['Orientation'] == 0) then
			gFlip = false
			gMirror = false
			Helper.StatusPrint ("No Flip or Mirror on image")
		elseif (gImageValues['Orientation'] == 1) then
			gFlip = true
			gMirror = false
			Helper.StatusPrint ("Image Flipped")
		elseif (gImageValues['Orientation'] == 2) then
			gFlip = false
			gMirror = true
			Helper.StatusPrint ("Image Mirrored")
		elseif (gImageValues['Orientation'] == 3) then
			gFlip = true
			gMirror = true
			Helper.StatusPrint ("Image Mirrored and Flipped")
		end
	end

	-- This is the last command for setting up so undo here.
	if (SETTINGUP) then
		SETTINGUP = false
	end

end

function RESPONSES.GetMotionServerSettings (data,responseCode, tHeaders)
	local server = string.find(data,'M_NotifyEn=')
	if (not server) then -- does not support server
		Helper.StatusPrint ('Camera does not support HTTP instant motion notification')
		return
	end

	-- set up server
	local ip = C4:GetControllerNetworkAddress ()
	local url = URL.BuildHTTP("MotionDetectionHttpNotify.cgi?M_NotifyEn=1&HttpUrl=http://" .. ip .. ":" .. gServerPort .. "/MotionEvent.cgi")
	local ticket = URL.GetRequest(url)
	if (not gDwell["MotionDetect"]) then
		gDwell["MotionDetect"] = Timer:Create("Motion Dwell", 5, "SECONDS", OnDwellExpired, false, "MotionDetect")
	end
	if (not gDetectionBinding["MotionDetect"]) then -- this contact has not been set up
		gDetectionState["MotionDetect"] = 'UNKNOWN'
		PersistData["detection"][110] = "Motion Detection"
		gDetectionBinding["MotionDetect"] = 110
		C4:AddDynamicBinding(110, "CONTROL", true, "Motion Detection", "CONTACT_SENSOR", false, false)
		C4:SendToProxy(110, "CLOSED", "", "NOTIFY")
	end
	Helper.StatusPrint ('Camera supports HTTP instant motion notification and this is enabled')
end

function RESPONSES.GetInputServerSettings (data,responseCode, tHeaders)
	local server = string.find(data,'I_NotifyEn=')
	if (not server) then -- does not support server
		Helper.StatusPrint ('Camera does not support HTTP input trigger notification')
		return
	end
	-- set up server
	local ip = C4:GetControllerNetworkAddress ()
	local url = URL.BuildHTTP("InputTriggerHttpNotify.cgi?I_NotifyEn=1&HttpUrl=http://" .. ip .. ":" .. gServerPort .. "/InputEvent.cgi")
	local ticket = URL.GetRequest(url)
	if (not gDwell["InputDetect"]) then
		gDwell["InputDetect"] = Timer:Create("Input Dwell", 5, "SECONDS", OnDwellExpired, false, "InputDetect")
	end
	if (not gDetectionBinding["InputDetect"]) then -- this contact has not been set up
		gDetectionState["InputDetect"] = 'UNKNOWN'
		gDetectionBinding["InputDetect"] = 111
		PersistData["detection"][111] = "Input Detection"
		C4:AddDynamicBinding(111, "CONTROL", true, "Input Detection", "CONTACT_SENSOR", false, false)
		C4:SendToProxy(111, "CLOSED", "", "NOTIFY")
	end
	Helper.StatusPrint ('Camera supports HTTP instant input trigger notification and this is enabled')
end


function RESPONSES.GetTamperServerSettings (data,responseCode, tHeaders)
	local server = string.match(data,'T_NotifyEn=(%d)')
	if (not server) then -- does not support server
		Helper.StatusPrint ('Camera does not support HTTP instant tamper notification')
		return
	end

	-- set up server
	local ip = C4:GetControllerNetworkAddress ()
	local url = URL.BuildHTTP("TamperingHttpNotify.cgi?T_NotifyEn=1&HttpUrl=http://" .. ip .. ":" .. gServerPort .. "/TamperEvent.cgi")
	local ticket = URL.GetRequest(url)
	if (not gDwell["TamperDetect"]) then
		gDwell["TamperDetect"] = Timer:Create("Tamper Dwell", 5, "SECONDS", OnDwellExpired, false, "TamperDetect")
	end
	if (not gDetectionBinding["TamperDetect"]) then -- this contact has not been set up
		gDetectionState["TamperDetect"] = 'UNKNOWN'
		gDetectionBinding["TamperDetect"] = 112
		PersistData["detection"][112] = "Tamper Detection"
		C4:AddDynamicBinding(112, "CONTROL", true, "Tamper Detection", "CONTACT_SENSOR", false, false)
		C4:SendToProxy(112, "CLOSED", "", "NOTIFY")
	end
	Helper.StatusPrint ('Camera supports HTTP instant tamper notification and this is enabled')

end

function RESPONSES.GetVideoServerSettings (data,responseCode, tHeaders)
	local server = string.find(data,'V_NotifyEn=')
	if (not server) then -- does not support server
		Helper.StatusPrint ('Camera does not support HTTP instant video notification')
		return
	end

	local ip = C4:GetControllerNetworkAddress ()
	local url = URL.BuildHTTP("VideoHttpNotify.cgi?V_NotifyEn=1&HttpUrl=http://" .. ip .. ":" .. gServerPort .. "/VideoEvent.cgi")
	local ticket = URL.GetRequest(url)
	if (not gDwell["VideoDetect"]) then
		gDwell["VideoDetect"] = Timer:Create("Video Dwell", 5, "SECONDS", OnDwellExpired, false, "VideoDetect")
	end
	if (not gDetectionBinding["VideoDetect"]) then -- this contact has not been set up
		gDetectionState["VideoDetect"] = 'UNKNOWN'
		gDetectionBinding["VideoDetect"] = 113
		PersistData["detection"][113] = "Video Detection"
		C4:AddDynamicBinding(113, "CONTROL", true, "Video Detection", "CONTACT_SENSOR", false, false)
		C4:SendToProxy(113, "CLOSED", "", "NOTIFY")
	end
	Helper.StatusPrint ('Camera supports HTTP instant video notification and this is enabled')
end

function RESPONSES.GetMotionSettings (data,responseCode, tHeaders)
	local motion = string.match(data,'md_enable=(%d)<br>')
	if (not motion) then
		return
	end

	local motionareas = tonumber(string.match(data,'md_area_num=(%d)<br>'))
	Helper.StatusPrint ('Camera supports ' .. motionareas .. ' motion detection areas')
	for i = 1, motionareas do
		motion = string.match(data,'md' .. i .. '_ena=(%d)<br>')
		if (SETTINGUP) then
			local binding = i + 100
			PersistData["detection"][binding] = "Motion Area " .. i
			C4:AddDynamicBinding(binding, "CONTROL", true, "Motion Area " .. i, "CONTACT_SENSOR", false, false)
			C4:SendToProxy(binding, "CLOSED", "", "NOTIFY")
			gDetectionState["Motion Area " .. i] = 'UNKNOWN'
			gContactBound[binding] = false
		end
		if (motion == '1') then
			Helper.StatusPrint ('Motion area ' .. i .. ' is enabled on the camera, virtual contact will switch')
		else
			Helper.StatusPrint ('WARN: Motion area ' .. i .. ' is not enabled on the camera, virtual contact will not switch')
			gError = true
		end
	end
	gMotionPoll = string.match(data,'Interval=(%d+)<br>')
	Helper.StatusPrint ("Motion polling frequency is ".. gMotionPoll .. " seconds")
end


function RESPONSES.GetNetworkSettings (data,responseCode, tHeaders)
	local dhcp = string.match(data,'Proto=(%d)<br>')
	if (not dhcp) then
		Helper.StatusPrint('No response to network settings query, check network settings and port')
		Helper.DriverInfo ('ERROR: No response from camera')
		gError = true
		return
	end
	gStatusTimer:StartTimer() -- if have made contact start to poll regularly
	if (dhcp == '1') then -- warn
		Helper.StatusPrint("WARN: Camera is set to DHCP - please set to Static IP or ensure MAC is reserved")
		gError = true
	end
	gDiscoveredHTTPPort = string.match(data,'Http_Port=(%d+)<br>')
	if (gDiscoveredHTTPPort == nil) then
		Helper.StatusPrint ('FATAL ERROR: NO HTTP Port found')
	else
		Helper.StatusPrint ('HTTP Port set to '.. gDiscoveredHTTPPort)
	end
	gDiscoveredRTSPPort = string.match(data,'Rtsp_Port=(%d+)<br>')
	if (gDiscoveredRTSPPort == nil) then
		Helper.StatusPrint ('FATAL ERROR: NO RTSP Port found')
	else
		Helper.StatusPrint ('RTSP Port set to '.. gDiscoveredRTSPPort)
	end
	if (string.match(data,'Rtsp_En=(%d+)<br>') ~= '1') then -- RTSP not enabled
		Helper.StatusPrint ('ERROR: Rtsp is not enabled - iOS and Android Apps will not show images.  Use "Enable RTSP" Action to correct.')
		gError = true
	end
	gDiscoveredPath1 = string.match(data,'Path_Live1=(.-)<br>')
	gDiscoveredPath2 = string.match(data,'Path_Live2=(.-)<br>')
	if (gDiscoveredPath1 and gDiscoveredPath2) then
		Helper.StatusPrint ('Video paths ' .. gDiscoveredPath1 .. ' and ' .. gDiscoveredPath2)
	else
		Helper.StatusPrint ('ERROR: No RTSP video path found')
	end
end

function RESPONSES.GetDateTimeSettings (data,responseCode, tHeaders)
	local localtime = string.match(data, "Localtime=(%d+)<br>")
	if (not localtime) then
		Helper.StatusPrint('No response to date and time query, check network settings and port')
		gError = true
		return
	end
	local difference = math.abs(localtime - os.time())
	if (difference > 60) then
		Helper.StatusPrint("Camera time differs from Control4 time by more than a minute, suggest synchronizing")
		gError = true
	else
		Helper.StatusPrint("Camera time differs from Control4 time by " .. difference .. " seconds" )
	end
	--LUA_ACTION.Time ()
end




function RESPONSES.GetMotion (data,responseCode, tHeaders)

	local motion = tonumber(string.match(data,'Motion0_Status=(%d)<br>'))
	if (motion) then
		if (motion == 0) then -- no motion
			Camera.SetMotion (1, 'CLOSED')
			Camera.SetMotion (2, 'CLOSED')
			Camera.SetMotion (3, 'CLOSED')

		elseif (motion == 1) then -- motion 1
			Camera.SetMotion (1, 'OPENED')
			Camera.SetMotion (2, 'CLOSED')
			Camera.SetMotion (3, 'CLOSED')

		elseif (motion == 2) then -- motion 2
			Camera.SetMotion (1, 'CLOSED')
			Camera.SetMotion (2, 'OPENED')
			Camera.SetMotion (3, 'CLOSED')

		elseif (motion == 3) then -- motion 1 & 2
			Camera.SetMotion (1, 'OPENED')
			Camera.SetMotion (2, 'OPENED')
			Camera.SetMotion (3, 'CLOSED')

		elseif (motion == 4) then -- motion 3
			Camera.SetMotion (1, 'CLOSED')
			Camera.SetMotion (2, 'CLOSED')
			Camera.SetMotion (3, 'OPENED')

		elseif (motion == 5) then -- motion 1 & 3
			Camera.SetMotion (1, 'OPENED')
			Camera.SetMotion (2, 'CLOSED')
			Camera.SetMotion (3, 'OPENED')

		elseif (motion == 6) then -- motion 2 & 3
			Camera.SetMotion (1, 'CLOSED')
			Camera.SetMotion (2, 'OPENED')
			Camera.SetMotion (3, 'OPENED')

		elseif (motion == 7) then -- motion 1,2 & 3
			Camera.SetMotion (1, 'OPENED')
			Camera.SetMotion (2, 'OPENED')
			Camera.SetMotion (3, 'OPENED')

		else
			Helper.StatusPrint ('Unknown motion respose '.. motion)
		end
	else
		Helper.StatusPrint ('No Motion Response Received')
	end
end

function RESPONSES.GetCameraVersion(data,responseCode, tHeaders)
	local ver = string.match(data,'Version=(.-)<br>')
	if (ver == nil) then
		Helper.StatusPrint ('No response to version, check network and port settings')
		gError = true
		return
	end
	Helper.StatusPrint ('Firmware Version ' .. ver)
	C4:UpdateProperty ('Firmware Version', ver)
end


function RESPONSES.RebootCamera (data,responseCode, tHeaders)
	Helper.DriverInfo('Camera rebooting - will query it in 60 seconds')
	Camera.ClearInfo()
	gRebootTimer:StartTimer()
end

function RESPONSES.SetAlarmDuration (data,responseCode, tHeaders)
		gMotionTimer:StartTimer(gMotionPoll, "SECONDS", true)
		OnMotionTimerExpired() -- Check immediately
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Helper functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function Helper.DriverInfo (info)
	C4:UpdateProperty ('Driver Information', info)
	print (os.date ('%x %X : ') .. info)
end
function Helper.Round (num, idp)
    local mult = 10^(idp or 0)
    if num >= 0 then return math.floor(num * mult + 0.5) / mult
    else return math.ceil(num * mult - 0.5) / mult end
end

function Helper.Print (data)
	if (type (data) == 'table') then
		for k, v in pairs (data) do print (k, v) end
	elseif (type (data) ~= 'nil') then
		print (type (data), data)
	else
		print ('nil value')
	end
end


function Helper.StatusPrint (data)
	if (gStatusPrint) then
		print(data)
	end
end

function Helper.ToBoolean(val)
	rval = false;

	if type(val) == "string" and (string.lower(val) == "true" or val == "1") then
		rval = true
	elseif type(val) == "number" and val ~= 0 then
		rval =  true
	elseif type(val) == "boolean" and val == true then
		rval = true
	end

	return rval
end
function Helper.Unlicenced ()
	--if (os.time() > EXPIRY) then
		--Helper.DriverInfo ('This driver has expired')
		--return true
	--else
		return false
	--end
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Camera functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function Camera.ClearInfo()
	C4:UpdateProperty ('Camera Model', '--')
	C4:UpdateProperty ('Camera Name', '--')
	C4:UpdateProperty ('Firmware Version', '--')
end

function login_a()
	local url = URL.BuildHTTP('API/Web/Login')
	local jdata = {version='1.0',
					data = {oem_type = 0, register_uid ='132abcd',remote_terminal_info='client' }
				  }
	local ticket = URL.PostRequest(url, C4:JsonEncode(jdata))
	gTickets[ticket] = 'Login_first'
end

function login()
	local t = C4:url()
	t:onHeaders( 
		function(transfer, responses)
			local lresp = responses[#responses]
			print("OnH: transfer succeeded (" .. #responses .. " responses received), lastresponse code: " .. lresp.code)
			for hdr,val in pairs(lresp.headers) do
			print("OnH: " .. hdr .. " = " .. val)
			end
		end
	)
	t:SetOptions({
		["fail_on_error"] = false,
		["timeout"] = 10,
		["connect_timeout"] = 5
		})
	local url = URL.BuildHTTP('API/Web/Login')
	local jdata = {version='1.0',
				   data = {oem_type = 0, register_uid ='132abcd',remote_terminal_info='client' }
 			 	  }
	t:Post(url,C4:JsonEncode(jdata),URL.AuthHeader())
end

function Camera.GetSettings()
	if (gStatusPrint) then
		Helper.DriverInfo ('Searching for Camera ------->')
	end
	gError = false
	gModel = nil
	
end
function Camera.TriggerAlert (detection)
	if (gDetectionState[detection]) then
		if (gDetectionState[detection] == 'CLOSED') then -- need to OPEN
			gDetectionState[detection] = 'OPENED'
			C4:SendToProxy (gDetectionBinding[detection], 'OPENED', '', 'NOTIFY')
			gDwell[detection]:StartTimer()
		else -- already OPENED so restart timer
			local interval = tonumber(gMotionPoll) or 10
			gDwell[detection]:StartTimer(interval, 'SECONDS')
		end
	else
		print ('No detection contact '..detection..' defined')
	end
	if (detection == "MotionDetect") then
		OnMotionTimerExpired()
	end
end

function Camera.GotoPreset(index)
	return "SetPTZ.cgi?CH=" .. gCameraID .. "&Dir=Point" .. index
end
function Camera.PTZControl(control)
	return "SetPTZ.cgi?CH=" .. gCameraID .. "&Dir=" .. control .. "&PTZSpeed=" .. gPtzIncrement
end
function Camera.Reboot ()
	return "Reboot.cgi?RebootNow"
end

function Camera.SetImageValue (parameter, value)
	return "SetImageValue.cgi?CH=" .. gCameraID .. "&" .. parameter .."=" .. value
end
function Camera.SetMotion (area, state)
	if (gDetectionState["Motion Area " .. area] ~= 'OPENED' and state == 'OPENED') then
		Dbg:Trace("Camera.SetMotion: Indicating Motion in area ".. area)
		gDetectionState["Motion Area " .. area] = 'OPENED'
		C4:SendToProxy (100+ area, 'OPENED', '', 'NOTIFY')
	elseif (gDetectionState["Motion Area " .. area] ~= 'CLOSED' and state == 'CLOSED') then
		Dbg:Trace("Camera.SetMotion: Indicating NoMotion in area ".. area)
		gDetectionState["Motion Area " .. area] = 'CLOSED'
		C4:SendToProxy (100+ area, 'CLOSED', '', 'NOTIFY')
	end
end
function Camera.SetOrientation (flip, mirror)
	local orientation = 'default'
	if (not mirror and not flip) then
		gFlip = false
		gMirror = false
		orientation = '0'
	elseif (flip and not mirror) then
		gFlip = true
		gMirror = false
		orientation = '1'
	elseif (not flip and mirror) then
		gFlip = false
		gMirror = true
		orientation = '2'
	elseif (flip and mirror) then
		gFlip = true
		gMirror = true
		orientation = '3'
	end

	return "SetImageValue.cgi?CH=" .. gCameraID .. "&Orientation=" .. orientation

end

function Camera.ServerSetUp ()
	local gControllerIP = C4:GetControllerNetworkAddress()
	local message = 'MotionDetectionHttpNotify.cgi?M_NotifyEn=1&HttpUrl=http://' .. gControllerIP .. ':' .. gServerPort ..'/'
	local url = URL.BuildHTTP(message)
	local ticket = URL.GetRequest(url)
	print (url)

end

---------------------------------------------------------------------
-- URL functions
---------------------------------------------------------------------
function URL.AuthHeader()

	local http_headers = {}

	if gAuthRequired then
		if gAauthType == "BASIC" then
			local http_header_field = gUsername .. ":" .. gPassword
			http_header_field = "Basic " .. C4:Base64Encode(http_header_field)
			http_headers["Authorization"] = http_header_field
		end
	end

	return http_headers
end

function URL.GetRequest(url)
	local l_url = url

	if gAuthRequired and gAauthType == "DIGEST" then
		l_url = string.sub(url, 1, 7) .. gUsername .. ":" .. gPassword .. "@" .. string.sub(url, 8)
	end

	return C4:urlGet(l_url, URL.AuthHeader())
end

function URL.PostRequest(url, data)
	local l_url = url

	if gAuthRequired and gAauthType == "DIGEST" then
		l_url = string.sub(url, 1, 7) .. gUsername .. ":" .. gPassword .. "@" .. string.sub(url, 8)
	end

	return C4:urlPost(l_url, data, URL.AuthHeader())
end

function URL.BuildHTTP(query_string)
	return "http://" .. gAddress .. ":" .. gHttpPort .. "/" .. query_string
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Common Driver Code
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	OnPropertyChanged
		Function called by Director when a property changes value.
	Parameters
		sProperty
			Name of property that has changed.
	Remarks
		The value of the property that has changed can be found with: Properties[sName]. Note
		that OnPropertyChanged is not called when the Property has been changed by the driver
		calling the UpdateProperty command, only when the Property is changed by the user from
		the Properties Page. This function is called by Director when a property changes value.
--]]
function OnBindingChanged (idBinding, strClass, bIsBound)
	-- No polling in this driver - will be dependent on server being set up
	--[[if ((idBinding > 100) and (idBinding < 200)) then -- binding to contact
		gContactBound[idBinding] = bIsBound
		-- now check if we should poll
		gCanPoll = false
		for k, v in pairs (gContactBound) do
			if (v == true) then
				gCanPoll = true
			end
		end

		if (gCanPoll == false) then -- stop any polling
			gMotionTimer:KillTimer()
		else
			if (gMotionPoll ~= 0) then -- not set this to do not poll in Properties
				gMotionTimer:StartTimer(gMotionPoll, 'SECONDS', true)
				OnMotionTimerExpired()
			end
		end
	end]]--
end

function OnNetworkBindingChanged (idBinding, bIsBound)
	print (idBinding,bIsBound)
		if (idBinding == 6001) then
		if (bIsBound == true) then
			NETWORK_BOUND = true
			gAddress = C4:GetBindingAddress (idBinding)
			local controllerip = C4:GetControllerNetworkAddress()
			C4:SendToProxy (DEFAULT_PROXY_BINDINGID, "ADDRESS_CHANGED", {ADDRESS = gAddress})
			local _, _, ip_range = string.find(gAddress, '(%d+.%d+.%d+).%d')
			local _, _, controller_range = string.find(controllerip, '(%d+.%d+.%d+).%d')
			if (ip_range ~= controller_range) then
				Helper.DriverInfo ('ERROR: Camera IP address not on same subnet as C4')
			else
				Helper.DriverInfo ('Attempting to connect to camera')
				gStatusPrint = true
				Camera.GetSettings ()
			end
		else
			C4:DestroyServer () -- Check this will reconnect OK
		end
	end
end

function OnPropertyChanged(sProperty)
	Dbg:Trace("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])
	
	local propertyValue = Properties[sProperty]

	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")
	print("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty] .. "trimprop: " .. trimmedProperty)
	-- if function exists then execute (non-stripped)
	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		ON_PROPERTY_CHANGED[sProperty](propertyValue)
		return
	-- elseif trimmed function exists then execute
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		ON_PROPERTY_CHANGED[trimmedProperty](propertyValue)
		return
	end
end

function ON_PROPERTY_CHANGED.DebugMode(propertyValue)
	gDebugTimer:KillTimer()
	Dbg:OutputPrint(propertyValue:find("Print") ~= nil)
	Dbg:OutputC4Log(propertyValue:find("Log") ~= nil)
	if (propertyValue == "Off") then return end
	gDebugTimer:StartTimer()
end

function ON_PROPERTY_CHANGED.DebugLevel(propertyValue)
	Dbg:SetLogLevel(tonumber(string.sub(propertyValue, 1, 1)))
end

function OnServerConnectionStatusChanged(nHandle, nPort, strStatus)
	if (strStatus == "ONLINE") then
		--print("Server Connection Established: " .. nHandle)
	else
		--print("Server Connection Disconnected: " .. nHandle)
	end
end

function OnServerDataIn(nHandle, strData)
	--print("Received Data on Handle: " .. nHandle .. ": " .. strData)
	if (string.find(strData, "Message=motion")) then
		Camera.TriggerAlert ('MotionDetect')
	elseif (string.find(strData, "Message=input")) then
		Camera.TriggerAlert ('InputDetect')
	elseif (string.find(strData, "Message=video")) then -- need to confirm
		Camera.TriggerAlert ('VideoDetect')
	elseif (string.find(strData, "Message=tampering")) then
		Camera.TriggerAlert ('TamperDetect')
	end
end
---------------------------------------------------------------------
-- ExecuteCommand Code
---------------------------------------------------------------------
--[[
	ExecuteCommand
		Function called by Director when a command is received for this DriverWorks driver.
		This includes commands created in Composer programming.
	Parameters
		sCommand
			Command to be sent
		tParams
			Lua table of parameters for the sent command
--]]
function ExecuteCommand(sCommand, tParams)

	if (Helper.Unlicenced ()) then return end
	Dbg:Trace("ExecuteCommand(" .. sCommand .. ")")
	Dbg:Info(tParams)

	-- Remove any spaces (trim the command)
	local trimmedCommand = string.gsub(sCommand, " ", "")

	-- if function exists then execute (non-stripped)
	if (EX_CMD[sCommand] ~= nil and type(EX_CMD[sCommand]) == "function") then
		EX_CMD[sCommand](tParams)
	-- elseif trimmed function exists then execute
	elseif (EX_CMD[trimmedCommand] ~= nil and type(EX_CMD[trimmedCommand]) == "function") then
		EX_CMD[trimmedCommand](tParams)
	-- handle the command
	elseif (EX_CMD[sCommand] ~= nil) then
		QueueCommand(EX_CMD[sCommand])
	else
		Dbg:Alert("ExecuteCommand: Unhandled command = " .. sCommand)
	end
end

--[[
	Define any functions of commands (EX_CMD.<command>) received from ExecuteCommand that need to be handled by the driver.
--]]

--[[
	EX_CMD.LUA_ACTION
		Function called for any actions executed by the user from the Actions Tab in Composer.
--]]
function EX_CMD.LUA_ACTION(tParams)
	if tParams ~= nil then
		for cmd,cmdv in pairs(tParams) do
			if cmd == "ACTION" then
				if (LUA_ACTION[cmdv] ~= nil) then
					LUA_ACTION[cmdv]()
				else
					Dbg:Alert("Undefined Action")
					Dbg:Alert("Key: " .. cmd .. " Value: " .. cmdv)
				end
			else
				Dbg:Alert("Undefined Command")
				Dbg:Alert("Key: " .. cmd .. " Value: " .. cmdv)
			end
		end
	end
end

function EX_CMD.AUTO_FOCUS() -- Not supported
end
function EX_CMD.FLIP_IMAGE()
	local url = URL.BuildHTTP(Camera.SetOrientation (true, gMirror))
	local ticket = URL.GetRequest(url)
end
function EX_CMD.UNFLIP_IMAGE()
	local url = URL.BuildHTTP(Camera.SetOrientation (false, gMirror))
	local ticket = URL.GetRequest(url)
end
function EX_CMD.MIRROR_IMAGE()
	local url = URL.BuildHTTP(Camera.SetOrientation (gFlip, true))
	local ticket = URL.GetRequest(url)
end
function EX_CMD.UNMIRROR_IMAGE()
	local url = URL.BuildHTTP(Camera.SetOrientation (gFlip, false))
	local ticket = URL.GetRequest(url)
end
function EX_CMD.REBOOT_CAMERA()
	local url = URL.BuildHTTP(Camera.Reboot ())
	local ticket = URL.GetRequest(url)
	gTickets[ticket] = 'RebootCamera'
end
function EX_CMD.SET_IMAGE_VALUE (tParams)
	if (gImageValues[tParams.Type] and (gImageValues[tParams.Type] ~= -1)) then -- this exists and is not -1
		local url = URL.BuildHTTP(Camera.SetImageValue (tParams.Type, tParams.Value))
		local ticket = URL.GetRequest(url)
	end
end

function EX_CMD.INC_IMAGE_VALUE (tParams)
	if (gImageValues[tParams.Type] and (gImageValues[tParams.Type] ~= -1)) then -- this exists and is not -1
		gImageValues[tParams.Type] = gImageValues[tParams.Type] + 1
		if (gImageValues[tParams.Type] > gImageValuesMax) then
			gImageValues[tParams.Type] = gImageValues[tParams.Type] - 1
		else
			local url = URL.BuildHTTP(Camera.SetImageValue (tParams.Type, gImageValues[tParams.Type]))
			local ticket = URL.GetRequest(url)
		end
	end
end

function EX_CMD.DEC_IMAGE_VALUE (tParams)
	if (gImageValues[tParams.Type] and (gImageValues[tParams.Type] ~= -1)) then -- this exists and is not -1
		gImageValues[tParams.Type] = gImageValues[tParams.Type] - 1
		if (gImageValues[tParams.Type] < gImageValuesMin) then
			gImageValues[tParams.Type] = gImageValues[tParams.Type] + 1
		else
			local url = URL.BuildHTTP(Camera.SetImageValue (tParams.Type, gImageValues[tParams.Type]))
			local ticket = URL.GetRequest(url)
		end
	end
end

function LUA_ACTION.DisplayGlobals()
	print ("Global Variables")
	print ("----------------------------")

	for k,v in pairs(_G) do	-- globals`
		if not (type(v) == "function") then
			--print(k .. ":  " .. tostring(v))
			if (string.find(k, "^g%L")  == 1) then
				print(k .. ":  " .. tostring(v))
				if (type(v) == "table") then
					PrintTable(v, "   ")
				end
			end
		end
	end

	print ("")
end

function LUA_ACTION.Flip ()
	local url = URL.BuildHTTP(Camera.SetOrientation (not gFlip, gMirror))
	local ticket = URL.GetRequest(url)
end
function LUA_ACTION.Mirror ()
	local url = URL.BuildHTTP(Camera.SetOrientation (gFlip, not gMirror))
	local ticket = URL.GetRequest(url)
end
function LUA_ACTION.Reboot ()
	gStatusPrint = true
	EX_CMD.REBOOT_CAMERA()
end
function LUA_ACTION.Reinitialize ()
	SETTINGUP = true
	gStatusPrint = true

	-- This is going to delete all the old detections
	for i = 1, 168, 1 do
		if(i < 10) then
			j = "10" .. i
		else
			j = "1" .. i
		end
		C4:RemoveDynamicBinding(j)
	end
	PersistData["detection"] = {}
	gDetectionState = {}
	gDetectionBinding = {}
	gImageValues = {}

	Camera.ClearInfo()
	Camera.GetSettings()
end
function LUA_ACTION.Report ()
	gStatusPrint = true
	Camera.ClearInfo()
	Camera.GetSettings()
end
function LUA_ACTION.Time ()
	Dbg:Trace("Time Sync ")
	local url = URL.BuildHTTP("SetTime.cgi?UseNtp=0&Localtime=" .. os.time())
	local ticket = URL.GetRequest(url)
end

function LUA_ACTION.RTSP ()
	Dbg:Trace("Enable RTSP")
	local url = URL.BuildHTTP("SetNetwork.cgi?Rtsp_En=1&Rtsp_Port=" .. gRtspPort)
	local ticket = URL.GetRequest(url)
end
function PrintTable(tValue, sIndent)
	sIndent = sIndent or "   "
	for k,v in pairs(tValue) do
		print(sIndent .. tostring(k) .. ":  " .. tostring(v))
		if (type(v) == "table") then
			PrintTable(v, sIndent .. "   ")
		end
	end
end

---------------------------------------------------------------------
-- ReceivedFromProxy Code
---------------------------------------------------------------------
--[[
	ReceivedFromProxy(idBinding, sCommand, tParams)
		Function called by Director when a proxy bound to the specified binding sends a
		BindMessage to the DriverWorks driver.

	Parameters
		idBinding
			Binding ID of the proxy that sent a BindMessage to the DriverWorks driver.
		sCommand
			Command that was sent
		tParams
			Lua table of received command parameters
--]]
function ReceivedFromProxy(idBinding, sCommand, tParams)
	if (Helper.Unlicenced ()) then return end
	if (sCommand ~= nil) then
		if(tParams == nil)		-- initial table variable if nil
			then tParams = {}
		end
		Dbg:Trace("ReceivedFromProxy(): " .. sCommand .. " on binding " .. idBinding .. "; Call Function " .. sCommand .. "()")
		Dbg:Info(tParams)

		if (PRX_CMD[sCommand]) ~= nil then
			PRX_CMD[sCommand](idBinding, tParams)
		else
			Dbg:Alert("ReceivedFromProxy: Unhandled command = " .. sCommand)
		end
	end
end

---------------------------------------------------------------------
-- Notification Code
---------------------------------------------------------------------
-- A notify with no parameters
function SendSimpleNotify(notifyText, ...)
	bindingID = select(1, ...) or DEFAULT_PROXY_BINDINGID
	C4:SendToProxy(bindingID, notifyText, {}, "NOTIFY")
end

---------------------------------------------------------------------
-- Initialization/Destructor Code
---------------------------------------------------------------------
--[[
	OnDriverInit
		Invoked by director when a driver is loaded. This API is provided for the driver developer to contain all of the driver
		objects that will require initialization.
--]]
function OnDriverInit()
	C4:ErrorLog("INIT_CODE: OnDriverInit()")
	-- Call all ON_DRIVER_EARLY_INIT functions.
	for k,v in pairs(ON_DRIVER_EARLY_INIT) do
		if (ON_DRIVER_EARLY_INIT[k] ~= nil and type(ON_DRIVER_EARLY_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_EARLY_INIT." .. k .. "()")
			ON_DRIVER_EARLY_INIT[k]()
		end
	end

	-- Call all ON_DRIVER_INIT functions
	for k,v in pairs(ON_DRIVER_INIT) do
		if (ON_DRIVER_INIT[k] ~= nil and type(ON_DRIVER_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_INIT." .. k .. "()")
			ON_DRIVER_INIT[k]()
		end
	end

	-- Fire OnPropertyChanged to set the initial Headers and other Property global sets, they'll change if Property is changed.
	for k,v in pairs(Properties) do
		OnPropertyChanged(k)
	end
end

--[[
	OnDriverUpdate
		Invoked by director when an update to a driver is requested. This request can occur either by adding a new version of a driver
		through the driver search list or right clicking on the driver and selecting "Update Driver" from within ComposerPro.
		Its purpose is to initialize all components of the driver that are reset during a driver update.
--]]
function OnDriverUpdate()
	C4:ErrorLog("INIT_CODE: OnDriverUpdate()")

	-- Call all ON_DRIVER_UPDATE functions
	for k,v in pairs(ON_DRIVER_UPDATE) do
		if (ON_DRIVER_UPDATE[k] ~= nil and type(ON_DRIVER_UPDATE[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_UPDATE." .. k .. "()")
			ON_DRIVER_UPDATE[k]()
		end
	end
end

--[[
	OnDriverLateInit
		Invoked by director after all drivers in the project have been loaded. This API is provided
		for the driver developer to contain all of the driver objects that will require initialization
		after all drivers in the project have been loaded.
--]]
function OnDriverLateInit()
	C4:ErrorLog("INIT_CODE: OnDriverLateInit()")

	-- Call all ON_DRIVER_LATEINIT functions
	for k,v in pairs(ON_DRIVER_LATEINIT) do
		if (ON_DRIVER_LATEINIT[k] ~= nil and type(ON_DRIVER_LATEINIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_LATEINIT." .. k .. "()")
			ON_DRIVER_LATEINIT[k]()
		end
	end
end


--[[
	OnDriverDestroyed
		Function called by Director when a driver is removed. Release things this driver has allocated such as timers.
--]]
function OnDriverDestroyed()
	C4:ErrorLog("INIT_CODE: OnDriverDestroyed()")
	-- Call all ON_DRIVER_DESTROYED functions
	for k,v in pairs(ON_DRIVER_DESTROYED) do
		if (ON_DRIVER_DESTROYED[k] ~= nil and type(ON_DRIVER_DESTROYED[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_DESTROYED." .. k .. "()")
			ON_DRIVER_DESTROYED[k]()
		end
	end
	C4:DestroyServer ()
end




--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Debug Logging Code
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
Log = {}

-- Create a Table with Logging functions
function Log:Create()

	-- table for logging functions
	local lt = {}

	lt._logLevel = 0
	lt._outputPrint = false
	lt._outputC4Log = false
	lt._logName =  "Set Log Name to display"

	function lt:SetLogLevel(level)
		self._logLevel = level
	end

	function lt:OutputPrint(value)
		self._outputPrint = value
	end

	function lt:OutputC4Log(value)
		self._outputC4Log = value
	end

	function lt:SetLogName(name)
		self._logName = name
	end

	function lt:Enabled()
		return (self._outputPrint or self._outputC4Log)
	end

	function lt:PrintTable(tValue, sIndent)
		if (type(tValue) == "table") then
			if (self._outputPrint) then
				for k,v in pairs(tValue) do
					print(sIndent .. tostring(k) .. ":  " .. tostring(v))
					if (type(v) == "table") then
						self:PrintTable(v, sIndent .. "   ")
					end
				end
			end

			if (self._outputC4Log) then
				for k,v in pairs(tValue) do
					C4:ErrorLog(self._logName .. ": " .. sIndent .. tostring(k) .. ":  " .. tostring(v))
					if (type(v) == "table") then
						self:PrintTable(v, sIndent .. "   ")
					end
				end
			end

		else
			if (self._outputPrint) then
				print (sIndent .. tValue)
			end

			if (self._outputC4Log) then
				C4:ErrorLog(self._logName .. ": " .. sIndent .. tValue)
			end
		end
	end

	function lt:Print(logLevel, sLogText)
		if (self._logLevel >= logLevel) then
			if (type(sLogText) == "table") then
				self:PrintTable(sLogText, "   ")
				return
			end

			if (self._outputPrint) then
				print (sLogText)
			end

			if (self._outputC4Log) then
				C4:ErrorLog(self._logName .. ": " .. sLogText)
			end
		end
	end

	function lt:Alert(strDebugText)
		self:Print(0, strDebugText)
	end

	function lt:Error(strDebugText)
		self:Print(1, strDebugText)
	end

	function lt:Warn(strDebugText)
		self:Print(2, strDebugText)
	end

	function lt:Info(strDebugText)
		self:Print(3, strDebugText)
	end

	function lt:Trace(strDebugText)
		self:Print(4, strDebugText)
	end

	function lt:Debug(strDebugText)
		self:Print(5, strDebugText)
	end

	return lt
end

function ON_DRIVER_EARLY_INIT.LogLib()
	-- Create and initialize debug logging
	Dbg = Log.Create()
	Dbg:SetLogName("xxoo Camera")
end

function ON_DRIVER_INIT.LogLib()
	-- Create Debug Timer
	gDebugTimer = Timer:Create("Debug", 45, "MINUTES", OnDebugTimerExpired)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Timer Expiry Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function OnDebugTimerExpired()
	Dbg:Warn("Turning Debug Mode Off (timer expired)")
	gDebugTimer:KillTimer()
	C4:UpdateProperty("Debug Mode", "Off")
	OnPropertyChanged("Debug Mode")
end

function OnDwellExpired (detection)
	Dbg:Trace("On Dwell Expired ".. detection)
	gDetectionState[detection] = 'CLOSED'
	C4:SendToProxy (gDetectionBinding[detection], 'CLOSED', '', 'NOTIFY')
	if (detection == "MotionDetect") then
		OnMotionTimerExpired()
	end
end
function OnResetInfoTimerExpired()
	Helper.DriverInfo ('Use Setup Report action to check camera settings')
end

function OnInitTimerExpired()
	Dbg:Trace("OnInitTimerExpired")
	if (gError) then
		Helper.DriverInfo ("Potential problems - please see Lua Tab")
		gResetInfoTimer:StartTimer()
	else
		Helper.DriverInfo (gModel .. " discovered")
	end
	gInitTimer:KillTimer()
	gStatusPrint = false
end

function OnMotionTimerExpired()
	Dbg:Trace("OnMotionTimerExpired")
	url = URL.BuildHTTP('GetMotionDetect.cgi')
	ticket = URL.GetRequest(url)
	gTickets[ticket] = 'GetMotion'
end

function OnPTZTimerExpired()
	local url = URL.BuildHTTP("goform/ptz?cmd=set&ptzparam=0&ptzcontrol=0&channel=" .. gCameraID)
	local ticket = URL.GetRequest(url)
end
function OnRebootTimerExpired()
	Camera.GetSettings ()
end
function OnSettingsTimerExpired()
	gStatusPrint = true
	Camera.GetSettings ()
end
function OnStatusTimerExpired()
	Camera.GetSettings ()
end
function OnSyncTimerExpired()
	LUA_ACTION.Time ()
end

---------------------------------------------------------------------
-- Timer Code
---------------------------------------------------------------------
Timer = {}

-- Create a Table with Timer functions
function Timer:Create(name, interval, units, Callback, repeating, Info)
	-- timers table
	local tt = {}

	tt._name = name
	tt._timerID = TimerLibGetNextTimerID()
	tt._interval = interval
	tt._units = units
	tt._repeating = repeating or false
	tt._Callback = Callback
	tt._info = Info or ""
	tt._id = 0

	function tt:StartTimer(...)
		self:KillTimer()

		-- optional parameters (interval, units, repeating)
		if ... then
			local interval = select(1, ...)
			local units = select(2, ...)
			local repeating = select(3, ...)

			self._interval = interval or self._interval
			self._units = units or self._units
			self._repeating = repeating or self._repeating
		end

		if (self._interval > 0) then
			Dbg:Trace("Starting Timer: " .. self._name)
			self._id = C4:AddTimer(self._interval, self._units, self._repeating)
		end
	end

	function tt:KillTimer()
		if (self._id) then
			self._id = C4:KillTimer(self._id)
		end
	end

	function tt:TimerStarted()
		return (self._id ~= 0)
	end

	function tt:TimerStopped()
		return not self:TimerStarted()
	end

	gTimerLibTimers[tt._timerID] = tt
	Dbg:Trace("Created timer " .. tt._name)

	return tt
end

function TimerLibGetNextTimerID()
	gTimerLibTimerCurID = gTimerLibTimerCurID + 1
	return gTimerLibTimerCurID
end

function ON_DRIVER_EARLY_INIT.TimerLib()
	gTimerLibTimers = {}
	gTimerLibTimerCurID = 0
end


function ON_DRIVER_DESTROYED.TimerLib()
	-- Kill open timers
	for k,v in pairs(gTimerLibTimers) do
		v:KillTimer()
	end
end

--[[
	OnTimerExpired
		Function called by Director when the specified Control4 timer expires.
	Parameters
		idTimer
			Timer ID of expired timer.
--]]
function OnTimerExpired(idTimer)
	for k,v in pairs(gTimerLibTimers) do
		if (idTimer == v._id) then
			if (v._Callback) then
				v._Callback(v._info)
			end
		end
	end
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Queue Code
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
Queue = {}

-- Create a Table with Queue functions
function Queue:Create()

	-- queue table
	local t = setmetatable(Queue, {
		__tostring = function(t)
			return t:_tostring()
		end
	})

	function t:_tostring()
		local tOutputString = {}

		table.insert(tOutputString, "--- Queue ---")
		table.insert(tOutputString, "  first = " .. tostring(self._et.first))
		table.insert(tOutputString, "  last = " .. tostring(self._et.last))
		table.insert(tOutputString, "  number in queue = " .. tostring(self._et.last - self._et.first + 1))
		table.insert(tOutputString, "  maximum size = " .. self._maxSize)
		table.insert(tOutputString, "  next value = " .. self:value())
		return table.concat(tOutputString, "\n")
	end

	-- entry table
	t._et = {first = 0, last = -1}
	t._maxSize = 0	-- no size limit

	-- push a value on the queue
	function t:push (value)
		local numItems = self._et.last - self._et.first + 1

		if ( (self._maxSize <= 0) or (numItems < self._maxSize) ) then
			local last = self._et.last + 1
			self._et.last = last
			self._et[last] = value
			--Dbg:Trace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		else
			-- if addToQueue == true then push value to queue
			if (self.OnMaxSizeReached()) then
				local last = self._et.last + 1
				self._et.last = last
				self._et[last] = value
				--Dbg:Trace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
			end
		end
	end

	function t:OnMaxSizeReached()
		--Dbg:Trace ("Max Size Reached - clear queue and push value to the queue (default).")
		local addToQueue = true

		t:clear()
		return (addToQueue)
	end

	-- pop a value from the queue
	function t:pop ()
		local first = self._et.first
		if first > self._et.last then
			--Dbg:Trace("Queue:pop(), queue is empty")
			return ""
		end
		local value = self._et[first]
		self._et[first] = nil        -- to allow garbage collection
		self._et.first = first + 1
		--Dbg:Trace ("Queue:pop(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		return value
	end

	-- clear queue
	function t:clear ()
		local first = self._et.first
		if first > self._et.last then
			--Dbg:Trace ("Queue:clear(), queue is empty")
			return ""
		end

		self._et = {first = 0, last = -1}
		--Dbg:Trace ("Queue:clear(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1))
		--Dbg:Trace (self._et)
	end

	-- return value of first item
	function t:value ()
		local first = self._et.first
		if (first > self._et.last) then
			return ""
		else
			return self._et[first]
		end
	end

	-- return queue's maximum size
	function t:MaxSize ()
		return self._maxSize
	end

	-- return queue's maximum size
	function t:SetMaxSize (size)
		self._maxSize = size
	end

	-- is queue empty?
	function t:empty ()
		-- print ("self._et.first = " .. tostring(self._et.first) .. ", self._et.last = " .. tostring(self._et.last))
		if (self._et.first > self._et.last) then
			return true
		else
			return false
		end
	end

	return t
end